import random


class NearestNeighbor:
    def __init__(self, graph, population):
        self.graph = graph
        self.population = population

    def get_init(self):
        random_paths = []
        list_vertices = self.graph.vertices

        for _ in range(self.population):
            random_start = random.sample(list_vertices, 1)[0]
            aux_list = list(list_vertices)
            aux_list.remove(random_start)
            random_list = [random_start]
            while len(aux_list) > 0:
                min_cost = float("Inf")
                next_node = None

                for i in aux_list:
                    if (
                        self.graph.edges_collection[(random_list[-1].idx, i.idx)].cost
                        < min_cost
                    ):
                        min_cost = self.graph.edges_collection[
                            (random_list[-1].idx, i.idx)
                        ].cost
                        next_node = i

                random_list.append(next_node)
                aux_list.remove(next_node)

            random_paths.append(random_list)

        return random_paths



import random
from unittest import TestCase
from tsp_solvers.tests.const import (
    GA_V10_COST,
    GA_V10_GENES,
    GA_V100_COST,
    GA_V100_GENES,
)
from tsp_solvers import Graph
from tsp_solvers import GeneticAlgorithm

# TODO: define tests for each method
#  use random.seed to be able to get the same results every time.


class GATestCase(TestCase):
    def setUp(self):
        super().setUp()
        random.seed(42)
        self.g = Graph()

    def tearDown(self):
        super().tearDown()
        pass

    def test_ga_10(self):
        self.g.create_graph_from_json("./tsp_solvers/tests/data/10v.json")
        ga = GeneticAlgorithm(self.g, 200, 100, 0.1, 120, "nearest")
        ga.run()
        self.assertListEqual(ga.best_genes, GA_V10_GENES)
        self.assertEqual(ga.best_cost, GA_V10_COST)

    def test_ga_100(self):
        self.g.create_graph_from_json("./tsp_solvers/tests/data/100v.json")
        ga = GeneticAlgorithm(self.g, 200, 100, 0.1, 120, "nearest")
        ga.run()
        self.assertListEqual(ga.best_genes, GA_V100_GENES)
        self.assertEqual(ga.best_cost, GA_V100_COST)



"""
This class implements common methods for all solving methods
"""


class BaseSolver:
    """ """

    def __init__(self):
        """ """
        self.time = None

    def get_time(self):
        """ """
        return self.time

    def show(self):
        """Has to be implemented in subclass"""
        raise NotImplemented

    def get_best(self):
        """Has to be implemented in subclass"""
        raise NotImplemented

    def run(self):
        """
        Has to be implemented in subclass
        """
        raise NotImplemented



from tsp_solvers.methods import (
    AntColonyOptimization,
    GeneticAlgorithm,
    LinearIntegerProgram,
    ParticleSwarmOptimization,
    SimulatedAnnealing,
    TwoOpt,
    Christofides,
)

from tsp_solvers.graph import Graph



"""
This file contains the main logic of the Genetic Algorithm method
"""
# TODO: study to implement part of the logic on C or C++ to speed up
# TODO: review convergence on algorithm as it always exists at 20% even with random initialization.

# Import from libraries
import datetime
import random
import sys

# Import from internal modules
from ..initializers import NearestNeighbor, RandomInitializer
from .base import BaseSolver


class Individual:
    """
    The genes are the solution (path)
    """

    def __init__(self, genes=None, cost=0, idx=0):
        if genes is None:
            genes = list()
        self.genes = genes
        self.cost = cost
        self.number_vertices = len(genes)
        self.idx = idx
        self._hash = self.__hash__()

    # TODO: change to property and setter methods
    def set_genes(self, genes):
        self.genes = list(genes)
        self.number_vertices = len(genes)

    def set_cost(self, cost):
        self.cost = cost

    def set_idx(self, idx):
        self.idx = idx

    def random_sub_solution(self):
        """"""
        start = random.randrange(0, self.number_vertices)
        end = start
        while (
            end == start
            or abs(end - start) == 1
            or abs(end - start) == len(self.genes) - 1
        ):
            end = random.randrange(0, self.number_vertices)
        if start > end:
            start, end = end, start
        return self.genes[start:end]

    def remove_vertices(self, sub_solution):
        """"""
        subset = set(sub_solution)
        self.genes = [gene for gene in self.genes if gene not in subset]

    def insert_sub_solution(self, sub_solution, x):
        """"""
        if x != -1:
            self.genes = self.genes[: x + 1] + sub_solution + self.genes[x + 1 :]
        else:
            self.genes = self.genes + sub_solution

    def mutate(self):
        """"""
        i = random.randrange(0, self.number_vertices)
        j = random.randrange(0, self.number_vertices)
        while i == j:
            j = random.randrange(0, self.number_vertices)

        aux = self.genes[i]
        self.genes[i] = self.genes[j]
        self.genes[j] = aux

    def __str__(self):
        return str(self.genes)


class GeneticAlgorithm(BaseSolver):
    def __init__(
        self,
        graph,
        max_generations,
        population_size,
        mutation_probability,
        max_time,
        init="random",
        plot=False,
        verbose=False,
    ):
        """ """
        super().__init__()
        self.individuals = list()
        self.graph = graph
        self.max_generations = max_generations
        self.population_size = population_size
        self.mutation_probability = mutation_probability
        self.mating_pool = []
        self.children = []
        self.id = 1
        self.max_time = max_time
        self.time = None
        self.verbose = verbose

        self.best_genes = None
        self.best_cost = float("inf")
        self.plot = plot
        if self.plot:
            if "matplotlib" not in sys.modules:
                raise ModuleNotFoundError(
                    "Matplotlib has to be installed to be able to plot!"
                )
        self.init = init
        self.initializer = None
        if self.init == "random":
            self.initializer = RandomInitializer(self.graph, self.population_size)
        elif self.init == "nearest":
            self.initializer = NearestNeighbor(self.graph, self.population_size)

        solutions = self.initializer.get_init()

        self.individuals = [
            Individual(
                genes=solution,
                cost=self.graph.get_solution_cost(solution),
                idx=counter + self.id,
            )
            for counter, solution in enumerate(solutions)
        ]
        self.id += len(solutions)

    def get_best(self):
        """"""
        best = self.individuals[0]
        for individual in self.individuals:
            if individual.cost < best.cost:
                best = individual
        print("\nGA:")
        print("Best solution: ", str(best.genes), "\t|\tcost: ", str(best.cost))

    def _update_best(self):
        """"""
        for individual in self.individuals:
            if individual.cost < self.best_cost:
                self.best_genes = individual.genes
                self.best_cost = individual.cost

    def _best_insertion(self, child, sub_solution):
        """"""
        start = sub_solution[0]
        end = sub_solution[-1]
        best_payoff = float("-inf")
        j = 0
        for i in range(0, len(child.genes) - 1):
            payoff = (
                self.graph.edges_collection[
                    (child.genes[i].idx, child.genes[i + 1].idx)
                ].cost
                - self.graph.edges_collection[(child.genes[i].idx, start.idx)].cost
                - self.graph.edges_collection[(end.idx, child.genes[i + 1].idx)].cost
            )

            if payoff > best_payoff:
                best_payoff = payoff
                j = i

        payoff = (
            self.graph.edges_collection[(child.genes[-1].idx, child.genes[0].idx)].cost
            - self.graph.edges_collection[(child.genes[-1].idx, start.idx)].cost
            - self.graph.edges_collection[(end.idx, child.genes[0].idx)].cost
        )

        if payoff > best_payoff:
            best_payoff = payoff
            j = -1

        return j

    def _cross_over(self):
        """"""
        self.children = []
        for n in range(0, self.population_size // 2):
            ind1 = random.choice(self.mating_pool)
            ind2 = random.choice(self.mating_pool)

            child = Individual()
            child.set_genes(ind1.genes)
            child.set_cost(ind1.cost)
            child.set_idx(self.id)
            self.id += 1
            sub_solution = ind2.random_sub_solution()

            child.remove_vertices(sub_solution)

            n = self._best_insertion(child, sub_solution)
            child.insert_sub_solution(sub_solution, n)
            child.cost = self.graph.get_solution_cost(child.genes)
            self.children.append(child)

    def _selection(self):
        """"""
        self.mating_pool = []
        iterator_list = list(self.individuals)
        while len(self.mating_pool) < self.population_size // 2:
            selected = min(iterator_list, key=lambda i: i.cost)
            iterator_list.remove(selected)
            self.mating_pool.append(selected)

    def _mutation(self):
        """"""
        for i in range(len(self.children)):
            prob = random.uniform(0, 1)
            if prob < self.mutation_probability:
                self.children[i].mutate()
                self.children[i].cost = self.graph.get_solution_cost(
                    self.children[i].genes
                )

    def _substitution(self):
        """"""
        new_population = []
        iterator_list = self.individuals + self.children

        while len(new_population) < self.population_size:
            selected = min(iterator_list, key=lambda i: i.cost)
            iterator_list.remove(selected)
            new_population.append(selected)

        self.individuals = new_population

    def _evaluate(self):
        """"""
        # TODO: review logic behind as this diversity is not working as intended and it stops prematurely with
        #  worse solutions than when this evaluation is not performed
        costs = [i.cost for i in self.individuals]
        mean_cost = sum(costs) / len(costs)
        min_cost = min(costs)
        if min_cost * 1.01 > mean_cost:
            if self.verbose:
                print("BREAK")
            return True
        else:
            return False

    def run(self):
        """"""
        start = datetime.datetime.utcnow()
        frac = 0.1
        plot = [10 * x for x in range(1, self.max_generations // 10 + 1)]
        initial_time = datetime.datetime.utcnow()
        for i in range(self.max_generations):
            if i / self.max_generations >= frac and self.verbose:
                print("Iteration ", str(i), " of ", str(self.max_generations))
                frac += 0.1
            # print("Generation: ", i + 1)
            self._selection()
            self._cross_over()
            self._mutation()
            self._substitution()
            self._update_best()

            if i >= self.max_generations // 5:
                if self._evaluate():
                    break

            if self.plot:
                if (i + 1) in plot:
                    filename = (
                        "plots/ga_"
                        + str(self.graph.number_vertices)
                        + "_"
                        + str(i + 1)
                        + ".png"
                    )
                    title = (
                        "Genetic algorithm. Iteration "
                        + str(i + 1)
                        + "\n Solution cost: "
                        + str(round(self.best_cost, 2))
                    )
                    self.graph.plot_solution(
                        self.best_genes,
                        pheromones=False,
                        filename=filename,
                        title=title,
                    )

            # self.get_best()
            if (datetime.datetime.utcnow() - initial_time).seconds > self.max_time:
                break

        self.time = datetime.datetime.utcnow() - start

        # print('FINISHED')

    def show(self):
        """"""
        print("\nSOLUTION:")
        print(
            "Individuals: "
            + str(self.population_size)
            + ". Iterations: "
            + str(self.max_generations)
        )
        for ind in self.individuals:
            print("Best solution: %s\t|\tcost: %d" % (str(ind.genes), ind.cost))

    def save(self, file):
        """"""
        best = self.individuals[0]
        for individual in self.individuals:
            if individual.cost < best.cost:
                best = individual
        text = (
            "GA. Best solution: "
            + str(best.genes)
            + "\t|\tCost: "
            + str(best.cost)
            + "\n"
        )
        with open(file, "a") as myfile:
            myfile.write(text)



"""

"""

import copy

from .base import BaseSolver


class Christofides(BaseSolver):
    # Only works on symmetric and completed graphs
    def __init__(self, graph, verbose=0):
        super().__init__()
        self.graph = graph
        self.verbose = verbose
        self.solution = None
        self.MST = None  # T
        self.odd_vertices = None  # O
        self.MWPM = None  # M

    def run(self):
        # Calculate minimum spanning tree (MST)
        # Subset O vertices of odd degree and calculate full graph
        # TODO: Construct minim weight perfect matching M
        # TODO: Unite graphs
        # TODO: Calculate Euler tour
        # TODO: Fix euler tour
        self._get_minimum_spanning_tree()
        self._get_odd_graph()
        self._get_minimum_weight_perfect_matching()
        pass

    def _get_minimum_spanning_tree(self):
        self.MST = copy.deepcopy(self.graph)
        self.MST.create_minimum_spanning_tree()

    def _get_odd_graph(self):
        self.odd_vertices = copy.deepcopy(self.MST)
        self.odd_vertices.create_full_odd_graph()
        self.odd_vertices.plot_edges()

    def _get_minimum_weight_perfect_matching(self):
        self.MWPM = copy.deepcopy(self.odd_vertices)
        self.MWPM.create_minimum_weight_perfect_matching()



"""
This file contains the main logic for the Ant Colony Optimization method
"""
# TODO: study to implement part of the logic on C or C++ to speed up

# Import from libraries
import datetime
import math
import random
import sys
from itertools import accumulate

# Import from internal modules
from .base import BaseSolver


class Ant:
    """
    The class for the individual ants
    """

    def __init__(self, alpha=0, beta=0, graph=None):
        """ """
        self.alpha = alpha
        self.beta = beta
        self.number_vertices = graph.number_vertices
        self.graph = graph
        self.solution = list()
        self.cost = 0
        self.unvisited_nodes = list(self.graph.vertices)

    def _select_node(self):
        """ """

        pheromone = [
            self.graph.edges_collection[
                (self.solution[-1].idx, unvisited.idx)
            ].pheromone
            for unvisited in self.unvisited_nodes
        ]

        # pheromone = []
        # for unvisited in self.unvisited_nodes:
        #     aux = self.graph.edges_collection[
        #         (self.solution[-1].idx, unvisited.idx)
        #     ].pheromone
        #     pheromone.append(aux)

        pheromone = list(map(lambda x: math.pow(x, self.alpha), pheromone))

        visibility = [
            self.graph.edges_collection[(self.solution[-1].idx, unvisited.idx)].cost
            for unvisited in self.unvisited_nodes
        ]

        visibility = list(map(lambda x: math.pow(1 / x, self.beta), visibility))

        denominator = sum([x * y for x, y in zip(pheromone, visibility)])
        cumulative_probability = list(
            accumulate([p * v / denominator for p, v in zip(pheromone, visibility)])
        )

        random_value = random.uniform(0, 1)

        added_idx = next(
            idx
            for idx, value in enumerate(cumulative_probability)
            if value >= random_value
        )

        added = self.unvisited_nodes[added_idx]
        self.unvisited_nodes.remove(added)
        return added

    def find_solution(self):
        """ """
        self.solution = random.sample(self.graph.vertices, 1)
        self.unvisited_nodes = list(self.graph.vertices)
        self.unvisited_nodes.remove(self.solution[0])
        while len(self.solution) < self.number_vertices:
            self.solution.append(self._select_node())
        return self.solution

    def get_cost(self):
        self.cost = self.graph.get_solution_cost(self.solution)
        return self.cost


class AntColonyOptimization(BaseSolver):
    def __init__(
        self,
        mode="AS",
        graph=None,
        iterations=100,
        population_size=10,
        alpha=1,
        beta=3,
        rho=0.1,
        pheromone_deposit=1,
        initial_pheromone=1,
        max_time=60,
        plot=False,
    ):
        super().__init__()
        self.mode = mode
        self.graph = graph
        self.iterations = iterations
        self.current_iteration = 0
        self.population_size = population_size
        self.alpha = alpha
        self.beta = beta
        self.rho = rho
        self.pheromone_deposit = pheromone_deposit
        self.initial_pheromone = initial_pheromone
        self.number_vertices = self.graph.number_vertices
        self.max_time = max_time
        self.plot = plot
        if self.plot:
            if "matplotlib" not in sys.modules:
                raise ModuleNotFoundError(
                    "Matplotlib has to be installed to be able to plot!"
                )

        self.time = None

        for edge in self.graph.edges_collection:
            self.graph.update_pheromone(edge, initial_pheromone)

        self.ants = [Ant(alpha, beta, self.graph) for _ in range(self.population_size)]

        self.best_solution = None
        self.best_cost = float("inf")

    def _add_pheromone(self, solution, cost, weight=1):
        pheromone_to_add = self.pheromone_deposit / cost
        for i in range(self.number_vertices):
            self.graph.edges_collection[
                (solution[i].idx, solution[(i + 1) % self.number_vertices].idx)
            ].pheromone = (
                self.graph.edges_collection[
                    (solution[i].idx, solution[(i + 1) % self.number_vertices].idx)
                ].pheromone
                + weight * pheromone_to_add
            )

    def get_best(self):
        best = self.ants[0]
        for ant in self.ants:
            if ant.cost < best.cost:
                best = ant
        print("\nACO:")
        print("Best solution: ", str(best.solution), "\t|\tcost: ", str(best.cost))

    def evaluate(self):
        costs = [i.cost for i in self.ants]
        mean_cost = sum(costs) / len(costs)
        min_cost = min(costs)

        if min_cost * 1.01 > mean_cost:
            print("BREAK")
            return True
        else:
            return False

    def _as(self):
        frac = 0.1
        initial_time = datetime.datetime.utcnow()
        for i in range(self.iterations):
            self.current_iteration = i
            if i / self.iterations >= frac:
                print("Iteration ", str(i), " of ", str(self.iterations))
                frac += 0.1

            if i != 0:
                for edge in self.graph.edges_collection:
                    self.graph.edges_collection[
                        edge
                    ].pheromone = self.graph.edges_collection[edge].pheromone * (
                        1 - self.rho
                    )

            for ant in self.ants:
                self._add_pheromone(ant.find_solution(), ant.get_cost(), 2)
                if ant.cost < self.best_cost:
                    self.best_solution = ant.solution
                    self.best_cost = ant.cost

            if i >= self.iterations // 5:
                if self.evaluate():
                    break

            if self.plot:
                self.plot_advancement()

            if (datetime.datetime.utcnow() - initial_time).seconds > self.max_time:
                break

    def run(self):
        start = datetime.datetime.utcnow()
        if self.mode == "AS":
            self._as()
        else:
            pass
        self.time = datetime.datetime.utcnow() - start

    def show(self):
        print("\nSOLUTION:")
        print(
            "Ants: "
            + str(self.population_size)
            + ". Iterations: "
            + str(self.iterations)
            + "\n"
        )
        for ant in self.ants:
            print("Best solution: %s\t|\t cost: %d" % (str(ant.solution), ant.cost))

    def save(self, file):
        best = self.ants[0]
        for ant in self.ants:
            if ant.cost < best.cost:
                best = ant
        text = (
            "ACO. Best solution: "
            + str(best.solution)
            + "\t|\tCost: "
            + str(best.cost)
            + "\n"
        )
        with open(file, "a") as f:
            f.write(text)

    def plot_advancement(self):
        plot = [10 * x for x in range(1, self.iterations // 10 + 1)]
        if (self.current_iteration + 1) in plot:
            filename = (
                "plots/aco_"
                + str(self.graph.number_vertices)
                + "_"
                + str(self.current_iteration + 1)
                + ".png"
            )
            title = (
                "Ant Colony Optimization. Iteration "
                + str(self.current_iteration + 1)
                + "\n Solution cost: "
                + str(round(self.best_cost, 2))
            )
            self.graph.plot_solution(
                self.best_solution,
                pheromones=True,
                filename=filename,
                title=title,
            )



"""
This file contains the Graph class that is used to represent the TSp problem.
This graph is constructed from nodes and edges
"""
# TODO: add more supported file types

# Import from libraries
import json
import random
import sys
from math import pow, sqrt
import traceback

# Matplotlib is not an required requirement
try:
    import matplotlib.pyplot as plt

    plt.ioff()
except ModuleNotFoundError:
    pass
except NameError:
    pass


class Edge:
    def __init__(self, data):
        self.origin = data.get("origin")
        self.destination = data.get("destination")
        self.cost = data.get("cost", 0)
        self.pheromone = data.get("pheromone", 0)
        self._hash = self.__hash__()

    def set_cost(self, cost):
        self.cost = cost

    def set_pheromone(self, pheromone):
        self.pheromone = pheromone

    def __hash__(self):
        return hash((self.origin.idx, self.destination.idx))

    def __eq__(self, other):
        if not isinstance(other, Edge):
            raise NotImplemented("The objects do not share the same class")
        return self._hash == other._hash

    def __repr__(self):
        return "Origin: {}, Destination: {}, Cost: {}".format(
            self.origin, self.destination, self.cost
        )


class Vertex:
    def __init__(self, data):
        self.idx = data.get("idx")
        self.x = data.get("x")
        self.y = data.get("y")
        self.degree = 0
        self._hash = self.__hash__()
        self.attributes_to_dict = ["idx", "x", "y"]

    def is_odd(self):
        return self.degree % 2 != 0

    def __hash__(self):
        return hash((self.idx, self.x, self.y))

    def __eq__(self, other):
        if isinstance(other, Vertex):
            return self._hash == other._hash
        elif isinstance(other, int):
            return self.idx == other
        else:
            raise NotImplemented("The objects do not share the same class")

    def __lt__(self, other):
        if isinstance(other, Vertex):
            return self._hash < other._hash
        elif isinstance(other, int):
            return self.idx < other
        else:
            raise NotImplemented("The objects do not share the same class")

    def __le__(self, other):
        if isinstance(other, Vertex):
            return self._hash <= other._hash
        elif isinstance(other, int):
            return self.idx <= other
        else:
            raise NotImplemented("The objects do not share the same class")

    def __repr__(self):
        return "Vertex: {}".format(self.idx)


class Graph:
    def __init__(self):
        self.data = None
        self.number_vertices = None
        self.edges = list()
        self.edges_collection = dict()
        self.vertices = list()
        self.vertices_collection = dict()
        # Used in Christofides approach
        self.components_groups = list()

    def calculate_cost(self, vertex_1, vertex_2):

        distance = sqrt(
            pow(
                vertex_1.x - vertex_2.x,
                2,
            )
            + pow(
                vertex_1.y - vertex_2.y,
                2,
            )
        )
        return distance

    def random_complete_graph(self, size):
        self.number_vertices = size

        for i in range(self.number_vertices):
            new = True
            while new:
                temp = {
                    "idx": i,
                    "x": random.randint(0, self.number_vertices * 2),
                    "y": random.randint(0, self.number_vertices * 2),
                }
                temp_vertex = Vertex(temp)
                repeat = False
                for vertex in self.vertices:
                    if vertex == temp_vertex:
                        repeat = True

                if not repeat:
                    new = False
            self.vertices.append(temp_vertex)
            self.vertices_collection[temp_vertex.idx] = temp_vertex

        self._create_edges()

    def create_graph_from_json(self, path: str):
        # TODO: add json schema validation here of kwargs so if data is passed no need to execute another method
        # TODO: add vertices_collection
        with open(path) as f:
            self.data = json.load(f)
        self.number_vertices = len(self.data.get("vertices"))
        self.vertices = [Vertex(vertex) for vertex in self.data.get("vertices")]
        self.vertices_collection = {vertex.idx: vertex for vertex in self.vertices}

        edges = self.data.get("edges", None)
        if edges is None:
            self._create_edges()

        else:
            # TODO: implement if the json has edges
            pass

    def create_graph_from_tsp(self, path: str):
        with open(path) as f:
            text = f.read()
            _, nodes = text.split("NODE_COORD_SECTION")
            nodes = nodes.split("\n")
            nodes = nodes[1:-1]
            nodes = [node.split(" ") for node in nodes]

        self.number_vertices = len(nodes)

        self.vertices = [
            Vertex({"idx": int(node[0]), "x": float(node[1]), "y": float(node[2])})
            for node in nodes
        ]

        self.vertices_collection = {vertex.idx: vertex for vertex in self.vertices}

        self._create_edges()

    def create_graph_from_db(self):
        pass

    def create_graph_from_schema(self, data):
        aux_vertices = list(set([d["n1"] for d in data] + [d["n2"] for d in data]))
        self.number_vertices = len(aux_vertices)
        self.vertices = [
            Vertex({"idx": vertex, "x": 0, "y": 0}) for vertex in aux_vertices
        ]

        self.vertices_collection = {vertex.idx: vertex for vertex in self.vertices}

        self.edges = [
            Edge(
                {
                    "origin": self.vertices_collection[d["n1"]],
                    "destination": self.vertices_collection[d["n2"]],
                    "cost": d["w"],
                }
            )
            for d in data
            if d["n1"] != d["n2"]
        ] + [
            Edge(
                {
                    "origin": self.vertices_collection[d["n2"]],
                    "destination": self.vertices_collection[d["n1"]],
                    "cost": d["w"],
                }
            )
            for d in data
            if d["n1"] != d["n2"]
        ]

        self.edges_collection = {
            (edge.origin.idx, edge.destination.idx): edge for edge in self.edges
        }

    def create_minimum_spanning_tree(self):
        # With Boruvka algorithm - kind of
        self._clean_edges()
        completed = False
        # First iteration
        for vertex in self.vertices:
            rest = {
                other: self.calculate_cost(vertex, other)
                for other in self.vertices
                if other != vertex
            }

            nearest = min(rest, key=rest.get)

            self.edges.append(
                Edge(
                    {
                        "origin": vertex,
                        "destination": nearest,
                        "cost": rest[nearest],
                    }
                )
            )
            self.edges_collection[(vertex.idx, nearest.idx)] = self.edges[-1]

        self._calculate_first_components()

        if len(self.components_groups) == 1:
            completed = True

        while not completed:
            # Function to check if we have only one component to exit loop

            for group1 in self.components_groups:
                rest = {}
                for group2 in self.components_groups:
                    if group1 != group2:
                        for origin in group1:
                            for destination in group2:
                                rest[origin, destination] = self.calculate_cost(
                                    origin, destination
                                )

                nearest = min(rest, key=rest.get)

                self.edges.append(
                    Edge(
                        {
                            "origin": nearest[0],
                            "destination": nearest[1],
                            "cost": rest[nearest],
                        }
                    )
                )

                self.edges_collection[(nearest[0].idx, nearest[1].idx)] = self.edges[-1]

                break

            self._update_components()

            if len(self.components_groups) == 1:
                completed = True
                self.plot_edges()

    def create_full_odd_graph(self):
        self._calculate_vertices_degrees()
        self._clean_edges()
        self._subset_odd_vertices()
        self._create_edges()

    def create_minimum_weight_perfect_matching(self):
        pass

    def get_random_paths(self, number):
        random_paths = []

        for i in range(number):
            random_list = random.sample(self.vertices, len(self.vertices))
            random_paths.append(random_list)

        return random_paths

    def get_solution_cost(self, path):
        total_cost = 0
        for i in range(self.number_vertices - 1):
            total_cost += self.edges_collection[(path[i].idx, path[i + 1].idx)].cost

        total_cost += self.edges_collection[
            (path[self.number_vertices - 1].idx, path[0].idx)
        ].cost
        return total_cost

    def update_pheromone(self, edge, pheromone):
        self.edges_collection[edge].pheromone = pheromone

    def save_graph_to_json(self, path):
        with open(path, "w") as f:
            json.dump(
                {
                    "vertices": [
                        {
                            key: value
                            for key, value in vertex.__dict__.items()
                            if key in vertex.attributes_to_dict
                        }
                        for vertex in self.vertices
                    ]
                },
                f,
            )

    def plot(self):
        x_coord = [vertex.x for vertex in self.vertices]
        y_coord = [vertex.y for vertex in self.vertices]
        if "matplotlib" not in sys.modules:
            raise ModuleNotFoundError(
                "Matplotlib has to be installed to be able to plot!"
            )
        plt.close()
        plt.scatter(x_coord, y_coord, c="#000000")
        plt.show()

    def plot_edges(self):
        x_coord = [vertex.x for vertex in self.vertices]
        y_coord = [vertex.y for vertex in self.vertices]
        if "matplotlib" not in sys.modules:
            raise ModuleNotFoundError(
                "Matplotlib has to be installed to be able to plot!"
            )
        plt.close()
        plt.scatter(x_coord, y_coord, c="#000000")

        for i in self.edges:
            x_values = [i.origin.x, i.destination.x]
            y_values = [i.origin.y, i.destination.y]
            plt.plot(x_values, y_values, c="#000000")
        plt.show()

    def plot_solution(self, path, pheromones=False, filename="none.png", title=""):
        if "matplotlib" not in sys.modules:
            raise ModuleNotFoundError(
                "Matplotlib has to be installed to be able to plot!"
            )
        plt.close()

        # if pheromones:
        #     min_pheromone = 100
        #     max_pheromone = 0
        #     for i in self.edges:
        #         if min_pheromone > self.edges[i].pheromone:
        #             min_pheromone = self.edges[i].pheromone
        #
        #         if max_pheromone < self.edges[i].pheromone:
        #             max_pheromone = self.edges[i].pheromone
        #
        #     pheromone_dict = {
        #         key: 0.0
        #         + (self.edges[key].pheromone - min_pheromone)
        #         * 0.4
        #         / (max_pheromone - min_pheromone)
        #         for key in self.edges
        #     }
        #
        #     for i in pheromone_dict:
        #         x_values = [
        #             self.vertices_coordinates[i[0]][0],
        #             self.vertices_coordinates[i[1]][0],
        #         ]
        #         y_values = [
        #             self.vertices_coordinates[i[0]][1],
        #             self.vertices_coordinates[i[1]][1],
        #         ]
        #         plt.plot(
        #             x_values,
        #             y_values,
        #             c="#FF0000",
        #             alpha=pheromone_dict[i],
        #             linewidth=3,
        #         )

        x_coord = [vertex.x for vertex in self.vertices]
        y_coord = [vertex.y for vertex in self.vertices]
        plt.scatter(x_coord, y_coord, c="#000000")
        for i in range(self.number_vertices - 1):
            x_values = [self.vertices[i].x, self.vertices[i + 1].x]
            y_values = [self.vertices[i].y, self.vertices[i + 1].y]
            plt.plot(x_values, y_values, c="#000000")

        x_values = [
            self.vertices[self.number_vertices - 1].x,
            self.vertices[0].x,
        ]
        y_values = [
            self.vertices[self.number_vertices - 1].y,
            self.vertices[0].y,
        ]
        plt.plot(x_values, y_values, c="#000000")

        plt.title(title)
        plt.savefig(filename)

    def _add_edge(self, vertex_1, vertex_2, cost=0.0, check=False):
        edge = Edge({"origin": vertex_1, "destination": vertex_2})
        if check:
            for created_edge in self.edges:
                if edge == created_edge:
                    return False

        edge.set_cost(cost)
        self.edges_collection[(vertex_1.idx, vertex_2.idx)] = edge
        self.edges.append(edge)

    def _calculate_vertices_degrees(self):
        # TODO: implement function that either calculates the degrees of all vertices or a given one
        temp_count = {k: 0 for k in self.vertices}
        counted = list()
        for edge in self.edges:
            if (edge.origin, edge.destination) in counted:
                continue
            temp_count[edge.origin] += 1
            temp_count[edge.destination] += 1
            counted.append((edge.destination, edge.origin))

        for key, value in temp_count.items():
            key.degree = value

    def _calculate_first_components(self):
        # TODO: review as it is not working properly.
        self.components_groups = list()
        for edge in self.edges:
            in_group = False
            if not self.components_groups:
                self.components_groups.append([edge.origin, edge.destination])
                continue

            for component in self.components_groups:

                if edge.origin in component or edge.destination in component:

                    if edge.origin in component and edge.destination in component:
                        in_group = True
                        break

                    elif edge.origin in component:
                        component.append(edge.destination)
                        in_group = True
                        break

                    elif edge.destination in component:
                        component.append(edge.origin)
                        in_group = True
                        break

            if not in_group:
                self.components_groups.append([edge.origin, edge.destination])

    def _update_components(self):
        origin = self.edges[-1].origin
        destination = self.edges[-1].destination

        origin_component = [c for c in self.components_groups if origin in c][0]
        destination_component = [c for c in self.components_groups if destination in c][
            0
        ]

        self.components_groups.remove(origin_component)
        self.components_groups.remove(destination_component)

        final_component = list(origin_component) + list(
            set(destination_component) - set(origin_component)
        )

        self.components_groups.append(final_component)

    def _clean_edges(self):
        self.edges = list()
        self.edges_collection = dict()

    def _subset_odd_vertices(self):
        self.vertices = [vertex for vertex in self.vertices if vertex.is_odd()]
        self.vertices_collection = {vertex.idx: vertex for vertex in self.vertices}

    def _create_edges(self):
        for i in self.vertices:
            for j in self.vertices:
                if i < j:
                    cost = self.calculate_cost(i, j)
                    self._add_edge(i, j, cost)
                    self._add_edge(j, i, cost)
